from PlugIns import PlugIn, PlugInGroup, PlugInContainer
import Acquisition
from ComputedAttribute import ComputedAttribute
from OFS.PropertySheets import PropertySheet
from Providers import Provider
from Globals import HTMLFile, default__class_init__

class SheetProviderGroup(PlugInGroup):
    """
    PlugInGroup that manages the more complex registration needs 
    of SheetProvider PlugIns/Containers
    """

    def _PlugInsChanged(self, container, plugins=()):

        """Update the registry of sheet names and XML namespaces"""

        sr = container.__sheet_registry__ = {}
        xr = container.__xmlns_registry__ = {}

        for plugin in plugins:
        
            names, namespaces = plugin.namesForRegistration(container)

            for n in names:
                if sr.has_key(n): sr[n].append(plugin)
                else:             sr[n]=[plugin]

            for n in namespaces: 
                if xr.has_key(n): xr[n].append(plugin)
                else:             xr[n]=[plugin]










    def manage_registries(self):
        """ """
        return (
            {'name':'Name', 
             'items':self.aq_parent.__sheet_registry__.items()
            },
            {'name':'XML Namespace', 
             'items':self.aq_parent.__xmlns_registry__.items()
            },
        )

    manage_below = HTMLFile('www/showSheetProviders', globals())

    __ac_permissions__ = (
        ('View management screens',
            ('manage_registries','manage_below')
        ),
    )


default__class_init__(SheetProviderGroup)




















psKey = 'ZPatterns.SheetProviders', 'PropertySheets'

class SheetProvider(Provider):
    """
    Sheet provider example class.  This sheet provider uses the 'slot'
    facility of Racks/Rackmountables to provide arbitrary propertysheets
    stored persistently.  All of its methods would need to be redefined
    for SheetProviders that implement fixed-schema and/or non-slot-based
    property sheets.
    """

    __plugin_kind__ = 'Property Sheet Provider'


    def namesForRegistration(self,container):
        """
Return a tuple (names,namespaces) where names is a sequence of
sheet names, and namespaces a sequence of XML namespaces, which this sheet
provider supports.  '*' may be included in either or both sequences, which
means the provider will be called if a more specific provider could not
be found for a given sheet.  This is mainly so that sheet providers can
potentially create propertysheets dynamically based on the requested name
or namespace.
        """
        return self.Sheet_Names, self.Sheet_Namespaces


    def _PropertySheetFor(self,client,name='',xmlns=''):
        """
        Return client's propertysheet matching name or xmlns, if available.
        Otherwise return None.
        """
        pslist = client._v_readableSlot.get(psKey,())
        for ps in pslist:
            if (name and ps.id==name) or (xmlns and ps.xml_namespace()==xmlns):
                return ps





    def _addPropertySheetFor(self,client,name='',xmlns=''):
        """
        Return client's new propertysheet matching name or xmlns, if possible.
        Otherwise return None.
        """
        ps=PropertySheet(name, {'xmlns':xmlns})
        pslist = client._v_readableSlot.get(psKey,())+(ps,)
        client._v_writeableSlot[psKey] = pslist
        return ps


    def _delPropertySheetFor(self,client,name='',xmlns=''):
        """
        Delete client's propertysheet matching name or xmlns, if present.
        Always returns None (unless an unexpected error occurs, of course)
        """
        pslist = client._v_readableSlot.get(psKey,())
        newlist = tuple(filter(
            lambda ps,name=name,xmlns=xmlns: 
                (not name or ps.id!=name) and 
                (not xmlns or ps.xml_namespace()!=xmlns),
            pslist
        ))
        if newlist!=pslist:
            client._v_writeableSlot[psKey] = newlist


    def _delAllSheetsFor(self,client):
        """Delete all of client's propertysheets"""
        if client._v_readableSlot.has_key(psKey):
            del client._v_writeableSlot[psKey]


    def _PropertySheetsFor(self, client):
        """Return a (possibly empty) list of PropertySheet objects for client"""
        # Note: this could include duplicate sheets if persistent sheets
        # are defined by more than one provider!
        return client._v_readableSlot.get(psKey,())



    # Class Metadata
    
    meta_type='Persistent Sheet Provider'
    
    Sheet_Names = ('*',)
    Sheet_Namespaces = ('*',)

    _properties=(
        {'id':'title', 'type': 'string', 'mode': 'w'},
        {'id':'Sheet_Names', 'type': 'tokens', 'mode': 'w'},
        {'id':'Sheet_Namespaces', 'type': 'lines', 'mode': 'w'},
    )

manage_addSheetProviderForm = HTMLFile('addSheetProvider', globals())

def manage_addSheetProvider(
    self, id, title='', REQUEST=None):
    
    """Add an Sheet Provider"""
    ob = SheetProvider(id, title)
    return self.Destination()._installPlugIn(ob, "Sheet Provider added", REQUEST)




























class AcquiredSheetProvider(SheetProvider):

    """Delegate providing of certain sheets to parent SPC"""
    
    def namesForRegistration(self,container):
        return (
            self._handledSheets(
                self.Sheet_Names, 
                container.aq_parent.__sheet_registry__.keys()
            ),
            self._handledSheets(
                self.Sheet_Namespaces,
                container.aq_parent.__xmlns_registry__.keys()
            )
        )
        
    def _handledSheets(self, filter, names):
        if '*' in filter:
            return names
        if '*' in names:
            return filter
        return filter(lambda x,f=filter: x in f, names)


    def _PropertySheetFor(self,client,name='',xmlns=''):
        return self.aq_parent.aq_parent._PropertySheetFor(client,name,xmlns)

    def _addPropertySheetFor(self,client,name='',xmlns=''):
        return self.aq_parent.aq_parent._addPropertySheetFor(client,name,xmlns)

    def _delPropertySheetFor(self,client,name='',xmlns=''):
        return self.aq_parent.aq_parent._delPropertySheetFor(client,name,xmlns)

    def _delAllSheetsFor(self,client):
        return self.aq_parent.aq_parent._delAllSheetsFor(client)
        
    def _PropertySheetsFor(self, client):
        # Note: this could include duplicate sheets if persistent sheets
        # are defined by more than one provider!
        return self.aq_parent.aq_parent._PropertySheetsFor(client)

    meta_type = 'Acquired Sheet Provider'
    
manage_addAcquiredSheetProviderForm = HTMLFile('addAcquiredSheetProvider', globals())

def manage_addAcquiredSheetProvider(
    self, id, title='', REQUEST=None):
    
    """Add an Acquired Sheet Provider"""
    ob = AcquiredSheetProvider(id, title)
    return self.Destination()._installPlugIn(ob, "Acquired Sheet Provider added", REQUEST)




class SheetProviderContainer:

    SheetProviders = SheetProviderGroup(
        'SheetProviders',
        ['Property Sheet Provider'],
        title = 'Sheet Providers',
        attr  = '__sheetProviders__'
    )

    __plugin_groups__ = (SheetProviders,)
    __sheetProviders__ = ()

    __sheet_registry__ = {}
    __xmlns_registry__ = {}


    def _PropertySheetFor(self, client, name='', xmlns=''):
        return self._delegateProviders(client,name,xmlns)


    def _addPropertySheetFor(self,client,name='',xmlns=''):
        """
Return client's new propertysheet matching name or xmlns, if possible.
Otherwise return None
        """
        return self._delegateProviders(client,name,xmlns,'_addPropertySheetFor')


    def _delPropertySheetFor(self,client,name='',xmlns=''):
        """
Delete client's propertysheet matching name or xmlns, if present.  Always
returns None (unless an unexpected error occurs, of course)
        """
        return self._delegateProviders(client,name,xmlns,'_delPropertySheetFor')


    def _delAllSheetsFor(self,client):
        """Delete all of client's propertysheets"""
        pass


    def _delegateProviders(self,client,name,xmlns,mname='_PropertySheetFor'):

        if xmlns:  k,d = xmlns, self.__xmlns_registry__
        elif name: k,d = name,  self.__sheet_registry__
        else:      return None

        self = self.aq_inner    # is this really necessary?

        for sp in d.get(k,[]) + d.get('*',[]):
            sp = sp.__of__(self)
            s = getattr(sp,mname)(client,name=name,xmlns=xmlns)
            if s is not None: return s

        parentMethod = getattr(self.aq_parent,'_delegateProviders',None)

        if parentMethod is not None:
            return parentMethod(client,name,xmlns,mname)


    def _PropertySheetsFor(self, client):
        """Return a (possibly empty) list of PropertySheet objects for client"""
        l = []
        for sp in self.__sheetProviders__:
            sp = sp.__of__(self)
            l.extend(list(sp._PropertySheetsFor(client)))

        parentMethod = getattr(self.aq_parent,'_PropertySheetsFor',None)

        if parentMethod is not None:
            l.extend(parentMethod(client))

        return l            


















